/* Generate some test data for Linear-regression tests. Outputs 4 columns of data (ascii-floating point, tab-separated, newlines, leading header with #) */

/* Copyright (C) Richard Neill 2011-2012, <maths at REMOVE.ME.richardneill.org>. This program is Free Software. You can
 * redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
 * either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
 * For the details, please see: http://www.gnu.org/licenses/gpl.html */

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <math.h>

#define  CHANNELS	4		/* Number of channels: 4 columns of data */
#define  SAMPLES_LONG	10000		/* Number of samples to create  (may be odd or even) */
#define  SAMPLES_MED	100		/* Medium and short files too */
#define  SAMPLES_SHORT	100
#define  PERIOD_US	5		/* Sample interval, in us. Scales the frequency. (cf 200kHz for NI4462) */

#define  PI	M_PI			/* 3.14... */

#define outprintf(...)		fprintf(outfile, __VA_ARGS__)

#define output(...)		fprintf (outfile, "%f\t%f\t%f\t%f\n", __VA_ARGS__);

FILE *outfile;

void printhelp(){
	fprintf (stderr, "This generates some sample data, suitable for testing linear regression algorithms.\n"
			 "The channels are 0 (linear slope), 1 (with noise), 2 (with noise and hum), 3 (exponential).\n"
			 "See the header of the generated file for the equations\n"
			 "USAGE:   genlinregtest   --mode  outfile.dat\n"
			 "MODES:   --long:      long dataset:   %d samples\n"
			 "         --medium:    medium dataset: %d samples\n"
			 "         --short:     short dataset:  %d samples\n"
			 "OUTPUT:  a simple format: 4 colums of ascii, tab-delimited floats, with #comment-lines.\n"
			 "         This format is, conveniently, parsed by Python's numpy.loadtxt(). \n"
			 "\n", SAMPLES_LONG, SAMPLES_MED, SAMPLES_SHORT);
}

void add_titles (char *comment, char *eqn0, char *eqn1, char *eqn2, char *eqn3){
 		fprintf (stderr, "Generating data: %s ...\n", comment);
 		outprintf ("# y0:\t\t%s\n", eqn0);
 		outprintf ("# y1:\t\t%s\n", eqn1);
 		outprintf ("# y2:\t\t%s\n", eqn2);
 		outprintf ("# y3:\t\t%s\n", eqn3);
}

/* Gaussian noise, mean 0, variance 1. */
double noise (){   /* See hawaiisim.c for details */
	int i;
	double z;
	z = 0;
	for (i=0; i<12; i++){
		z += ((double)rand() / RAND_MAX);   /* uniform [0,1] */
	}
	return (z - 6);
}

int main ( int argc, char **argv ) {
	char *outfile_filename, *type, *comment, *eqn0, *eqn1, *eqn2, *eqn3;
	double t, y0,y1,y2,y3, T;		/* t = time, y_n = output, T = total run time. */
	int samples, i;
	srand (time (NULL));			/* Seed random number generator (omit this to make the "random" data reproducable). */

	/* Process options */
	if (argc < 2){
		fprintf ( stderr, "Syntax: %s --type outfile.dat    #-h for help.\n", argv[0] );
		exit (1);
	}
	if ((!strcmp (argv[1], "-h")) || (!strcmp (argv[1], "--help"))){
		printhelp();
		exit (1);
	}
	if ( argc != 3 ) {
		fprintf ( stderr, "Syntax: %s --type outfile.dat    #-h for help.\n", argv[0] );
		exit (1);
	}
	type = argv[1];
	outfile_filename = argv[2];

	if ( ( outfile = fopen ( outfile_filename, "w" ) ) == NULL ) {
		fprintf ( stderr, "Could not open %s for writing: %s\n", outfile_filename, strerror ( errno ) );
		exit ( 1 );
	}

	/* Run length */
	if (!strcmp (type, "--long")){
		samples = SAMPLES_LONG;
	}else if (!strcmp (type, "--medium")){
		samples = SAMPLES_MED;
	}else if (!strcmp (type, "--short")){
		samples = SAMPLES_SHORT;
	}else{
		fprintf (stderr, "Sorry, format %s not recognised. Use -h for list.\n", type);
		exit(1);
	}

	T =  (double)(PERIOD_US * samples)/1e6;  /* total run time for all the samples */


	/* Output header */
	outprintf ("# This is some sample data for LinReg tests:\n");
	outprintf ("# channels:\t%d\n", 		CHANNELS);
	outprintf ("# samples:\t%d\n",  		samples);
	outprintf ("# sample_interval_us:\t%d\n", 	PERIOD_US);
	outprintf ("# sampling_freq_hz:\t%f\n", 	((double)1e6/PERIOD_US));
	outprintf ("# length_s:\t%f\n", T);


	/* Generate some data, and output it. */
	comment = "basic linear decrease, with noise/hum/exponential";
	eqn0   =  "y0 = A + B.t                              A=1, B=20";
	eqn1   =  "y1 = A + B.t + C.noise()                  A=2, B=-5,  C=0.01";
	eqn2   =  "y2 = A + B.t + C.noise() + D.cos(w t)     A=3, B=-10, C=0.1,  D=0.1, f=50 Hz";
	eqn3   =  "y3 = A.exp(- B.t) + C.noise()             A=4, B=-15, C=0.1";

	add_titles (comment, eqn0, eqn1, eqn2, eqn3);

	for (i = 0 ; i < samples; i++){
		t = (double)(i * PERIOD_US) / 1e6;

		y0 = 1 + (-20 * t);
		y1 = 2 + (-5  * t)  +      0.01 * noise();
		y2 = 3 + (-10  * t)  +      0.1 * noise()  + 0.1 * cos (50 * 2 * PI * t);
		y3 = 4 * exp (-15  * t) +   0.1 * noise();

		output (y0, y1, y2, y3);
	}


	/* End */
	fclose ( outfile );
	return (0);
}
