#!/usr/bin/python

#This program reads in a data-file, and plots it, together with its linear regression best-fit line. It prints the statistics.
#If the file has several columns, we can choose any one or the sum, or all of them.

#Copyright (C) Richard Neill 2011-2012, <maths at REMOVE.ME.richardneill.org>. This program is Free Software. You can
#redistribute and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation,
#either version 3 of the License, or (at your option) any later version. There is NO WARRANTY, neither express nor implied.
#For the details, please see: http://www.gnu.org/licenses/gpl.html

#TODO: Would be nice to allow error-bars to be shown in the case of "-c all -s". See: FIXME-34626351 below.
#TODO: Would be nice if the error-bars could be made less opaque, when they cover most of the underlying graph,
#TODO: Annotate the graph with the regression-line equation


## MATHS --------------------------------------------------

#Mathematics:
# We are doing Linear Regression with one dependent and one independent variable, so use Ordinary Least Squares method of Simple Linear regression.
# See:  http://mathworld.wolfram.com/LeastSquaresFitting.html
#       http://en.wikipedia.org/wiki/Linear_regression
#       http://en.wikipedia.org/wiki/Simple_linear_regression

#Note on notation. In Ascii, I use _ for subscript, eg  y_i.   Also, for clarity, I use  a and A, b and B etc, rather than alpha-hat and  alpha etc.
#Beware of the "S-notation": SS_xx, S_xx, s_x  mean, respectively: Sigma(x - x_bar)^2,  Sigma(x^2), std_dev(x)

#Our underlying data is assumed to be linear with random,gaussian noise:
#	y_i   = A + B x_i + E_i			n samples, E_i has std-dev sigma.

#Unbiased estimators of A,B  are a,b given by
#	b = Cov(x,y) / Var(x)
#       a = Mean(y) - b * Mean(x)

#Errors in a,b are as follows.  Error in mean is zero ("unbiased"). Std errors are:
#	SE(b)^2 = n.s^2 / (n.S_xx - (S_x)^2)
#       SE(a)^2 = SE(b)^2 * S_xx / n
#where  s^2 =  (1/(n*(n-2))) * (n.S_yy - (S_y)^2 - b^2(n.S_xx - (S_x)^2) ) 	# s is an estimator for the std-dev of the Noise, E_i

#Product moment correlation coefficient is rho (for population), estimated by r (for sample, usually written r-hat). Range is [-1 ... 0 ... 1]
#	r = ( n.S_xy - (S_x)^2 ) / Sqrt [ (n.S_xx - (S_x)^2) * (n.S_yy - (S_y)^2) ]


#Actually doing it:
# 1. Calculate 5 sums:
#    S_x   =  Sum(x)
#    S_y   =  Sum(y)
#    S_xx  =  Sum(x^2)
#    S_yy  =  Sum(y^2)
#    S_xy  =  Sum(xy)

# 2. Estimators are:
#    b     =  [ n S_xy -  S_x S_y ] /  [ n S_xx - (Sx)^2 ]		#b-hat
#    a     =  S_y / n  -  b S_x / n					#a-hat
#    r     = ( (n * S_xy) - (S_x * S_y) ) / Sqrt( (n * S_xx - (S_x * S_x)) * (n * S_yy - (S_y * S_y)) )		#r-hat
#    s     = Sqrt ( (1.0 / (n * (n-2))) * ( n * S_yy - (S_y * S_y) - (b * b * (n * S_xx - (S_x * S_x)) ) ))	#sigma-hat
#    SE(b) = Sqrt ( ( n * (s * s) ) / ( n * S_xx - (S_x * S_x) ) ) 						#std err in b-hat
#    SE(a) = Sqrt ( (se_b * se_b) * S_xx / n )									#std err in a-hat

#Physics:  slope = rate of photon arrival; so Delta(y) from start to end =  total photons.
#          BUT,  b * Delta(x) is actually a *better* estimate of the star brightness than Delta(y).

#Note 1: Don't use Python's polyfit() function - that calculates using total-least-squares, rather than vertical-least-squares. Could use lstsq() though.
#Note 2: Numerical hygiene: instead of Sqrt(), use Sqrt(Abs()) in case of tiny negative numbers. Also, avoid accidental integer division!
#Note 3: The parameters we get aren't necessarily a complete picture of the data: see "Anscombe's Quartet"
#Note 4: this OLS measurement is better than the CDS alternative. See proofs/linreg.py


## END ----------------------------------------------------


#Begin
from __future__ import division		#Needed, as this may run under Python 2.x
import getopt, sys, os, string
import numpy as numpy
from matplotlib import *
from pylab import *
from textwrap import *


#Help
def usage():
	print 	"linregplot:  plot a data-file, and its simple-linear-regression best-fit. Show summary information.\n"
	print 	"USAGE:    " + str(os.path.basename(sys.argv[0])) + "  [OPTS]  infile.dat\n"
	print	"OPTIONS:  -f, --samplefreq  FREQ        define the sampling frequency in Hz. [default: 1.0]."
	print   "          -c, --column      n,sum,all   use Nth column's data (N is 0-based). [default: all].\n"
	print   "          -t, --title       TITLE       override plot title"
	print   "          -x, --xlabel      LABEL       override x-axis label"
	print   "          -y, --xlabel      LABEL       override y-axis label"
	print   "          -u, --ulabel      L,A,BE,LS   override legend labels\n"
	print   "          -s, --separate                plot raw and regession graphs separately. [default: same axes]."
	print   "          -o, --outplot     FILE.png    save main plot as FILE.png/.jpg, \"auto\" names it after infile.\n"
	print   "          -n, --noplot                  don't draw the plot window on screen (useful for testing/scripting)."
	print   "          -h, --help                    show help\n"
	print   "INPUT:    Input file, infile.dat ('-' for stdin) should be columns of (int/float) data, separated by"
	print   "          whitespace, with optional #comments, i.e. the standard form for numpy.loadfile(). cf. genffttest."
	print   "          (compressed files (.gz/.bz2) can also be read transparently, no need to uncompress first.)\n"
	print   "NOTE 1:   Read this source for an explanation of the simple least-squares linear-regression maths.\n"
	print   "NOTE 2:   Error-bars (in grey) are only shown for single-column plots, i.e. except when \"-c all\"."
	print   ""
	return

#Initialise
plotdims=(12, 9); plotdpi=80; 	#Sets initial size of plot on screen, and the dimensions of the saved image.  DPI affects font-size.
samplerate = plotseparate = noplot = title = xlabel = ylabel = ulabel = outplot = False
column = -1
info_txt = '';

#Parse args
try:
	opts, args = getopt.gnu_getopt(sys.argv[1:], "c:f:hnst:x:y:u:o:",
	["column=", "samplefreq=", "help", "noplot", "separate", "title=", "xlabel=", "ylabel=", "ulabel=", "outplot="])
except getopt.GetoptError, err:
	print str(err) + ": use -h for help." ; sys.exit(2)

for o, a in opts:
	if o in ("-h", "--help"):
		usage(); exit(0);
	elif o in ("-f", "--samplefreq"):
		samplerate = float(a)
	elif o in ("-c", "--column"):
		if (a == "all") or (a == "sum"):
			column = a
		else:
			column = int(a)
	elif o in ("-s", "--separate"):
		plotseparate = True
	elif o in ("-n", "--noplot"):
		noplot = True
	elif o in ("-t", "--title"):
		title = fill(a, 80)       #wrap long lines if necessary (otherwise they get truncated)
	elif o in ("-x", "--xlabel"):
		xlabel = a
	elif o in ("-y", "--ylabel"):
		ylabel = a
	elif o in ("-u", "--ulabel"):
		ulabel = a
	elif o in ("-o", "--outplot"):
		outplot = a
	else:
		assert False, "unhandled option"

if ((len(args) == 0) and ('-' in sys.argv)):		#Ugly hack for older versions.
		infile = sys.stdin
		infilename = "<stdin>"
		if (outplot == "auto"):
			outplot = "stdin.png"
elif (len(args) != 1):
		print "Error: One (non-option) argument is required: input-filename."
		exit (2)
else:
	infile = args[0]
	infilename = infile

	if (infile == "-"):
		infile = sys.stdin
		infilename = "<stdin>"
		if (outplot == "auto"):
			outplot = "stdin.png"
	else:
		if (outplot == "auto"):
			outplot = os.path.splitext(os.path.basename(infilename))[0] + ".png"


#Load data from file.
try:
	data_in = numpy.loadtxt (infile);				#numpy.loadtxt really is all we need! It even handles comments :-)
except IOError, err:							#the function is very flexible, though the defaults suffice here.
	print "Error reading datafile " + infilename + ": " + str(err)
	exit (1);
except ValueError, err:
	print "Error processing malformed datafile " + infilename + ": " + str(err)
	exit (1);


#Get size of data file, sanity check it.
if (len(data_in.shape) == 1):		#if it is a single column (1d-array),  convert to a 2d-array with 1 column.
	data_in = data_in[:,newaxis]
num_cols = len(data_in[0, ]);
num_rows = len(data_in);
if (num_cols == 0) or (num_cols > 20) or (num_rows < 4) or (num_rows > 1E7):
	print "Data size is out of range. File contains "  + str(num_rows) + " rows of " + str(num_cols) + " columns."
	exit (1)


#Extract the relevant column.
if (column == -1):			#Default: All columns.
	column = "all"
	if (num_cols == 1):		#..but if the file only has 1 column, don't make it an array (for clarity).
		column = 0
if (column == "sum"):			#Sum the columns
	data = data_in [ : , 0]
	for col in range (1, num_cols):
		data += data_in [ : , col]
	legend = ["sum (0:"+str(num_cols -1)+")"]
elif (column == "all"):			#All columns (in parallel).
	data = data_in
	legend = ["col 0"]
	for col in range (1, num_cols):
		legend.append ("col " +str(col))
elif (column < num_cols):
	data = data_in [ : , column]	#This pulls out a 1-d array, from a 2-d array.
	legend = ["col "+str(column)]
else:
	assert False, "invalid column specified: " + str(column)

#Sampling number and rate.
num_samples = len(data)			#This many samples.
if (samplerate == False):		#May be given as arg, else 1.0 (float)
	samplerate = 1.0
times =	arange (0, num_samples) * (1.0/samplerate) #Create time-series for raw-data x-axis, instead of just showing sample-number
duration = num_samples / samplerate


#Calculate the parameters. (see above). Works for 1 col or n cols.
x 	= times.repeat(size(data)/shape(data)[0]).reshape(shape(data))
y	= data
n       = len(y)

S_x 	= sum(x, axis=0)		#Get a 1 or n-element result.
S_xx    = sum((x * x), axis=0)
S_y     = sum(y, axis=0)
S_yy    = sum((y * y), axis=0)
S_xy    = sum((x * y), axis=0)

b       = ( n * S_xy -  S_x * S_y ) /  ( n * S_xx - (S_x * S_x) )	#b-hat, estimator for gradient.
a       = ( S_y / n ) - ( b * S_x / n)					#a-hat, estimator for y-intercept.

if (n <= 2):
	print "Error, n is "+str(n)+" which is less than 3. Cannot calculate most estimators. Got y = a + b.x, where a is "+str(a)+" and b is "+str(b)+"."
	exit (1)

s = sqrt(abs( (1.0 / (n * (n-2))) * ( n * S_yy - (S_y * S_y) - (b * b * (n * S_xx - (S_x * S_x)) ) )))	#sigma-hat, (estimator of std-dev of noise)
se_b =  sqrt (abs ( ( n * (s * s) ) / ( n * S_xx - (S_x * S_x) ) ) )	#std err in b-hat
se_a = sqrt ( (se_b * se_b) * S_xx / n )				#std err in a-hat
r = ( (n * S_xy) - (S_x * S_y) ) / sqrt(abs( (n * S_xx - (S_x * S_x)) * (n * S_yy - (S_y * S_y)) ))	#r-hat, estimator for Pearson's product-moment-correlation-coefficient.
delta_y = y[n-1] - y[0]
b_delta_x = b * (times[n-1] - times[0])
#NB the "abs()" and "1.0/" are both necessary: the abs for numerical hygiene with tiny negative numbers; the 1.0 to make sure we don't do integer division!
#print "Debug: S_x=", S_x, " S_y=", S_y, " S_xx=", S_xx, " S_yy=", S_yy, " S_xy=", S_xy, " n=",  n, " a=",  a, " b=",  b, " s=",  s, " se_a=",  se_a, " se_b=",  se_b, " r=",  r, " delta-y=",  delta_y, " b.delta-x=",  b_delta_x


#Generate the regression-line's data.
regression = a + b * x


#Plot the data. Time is in seconds (implied by units of -f). Arbitrarily choose "Volts" to make it obvious what this is.
#Plot is very good at handling multiple lines, if data is 2-d array.
fig = plt.figure(figsize=plotdims, dpi=plotdpi);
plt.xlabel("Time (s)" if (xlabel == False) else xlabel)
ax1 = fig.add_subplot(111)
plt.ylabel("Signal (V)" if (ylabel == False) else ylabel)
lines = plt.plot (x, y, marker='.')
ax1_ylim = plt.ylim()
plt.grid()
if (num_cols > 1):
	plt.legend(legend)
if (ulabel != False):
	ulabels = string.split (ulabel, ",", (num_cols - 1) )	#comma separated.
	plt.legend (ulabels)
	if ((column < num_cols) and (len(ulabels) >= column)):  #select nth label if possible.
		plt.legend ([ulabels[column]])

if (plotseparate):	#Show regression line on separate graph. Add errorbars if we can (there are usually so many they appear blurred together).
	plt.title ("Raw signal data [ " + infilename + ", column " + str(column) + " ]" if (title == False) else title)

	fig = plt.figure(figsize=plotdims, dpi=plotdpi);
	plt.title ("Linear Regression (Least Squares)" if (title == False) else title)
	plt.xlabel("Time (s)" if (xlabel == False) else xlabel)
	plt.ylabel("Signal (V)" if (ylabel == False) else ylabel)
	if (column == "all"):
		plt.plot (times, regression, linestyle="--") #without error-bars: we can't have them, because plt.errorbar(yerr=) can't be a list.  [FIXME-34626351]
	else:
		plt.errorbar (times, regression, yerr=s, linestyle="--", ecolor='lightgrey') # marker='o', markevery=(0,n/20))
	plt.grid()
	if (num_cols > 1):
		plt.legend(legend)
	if (ulabel != False):
		ulabels = string.split (ulabel, ",", (num_cols - 1) )	#comma separated.
		plt.legend (ulabels)
		if ((column < num_cols) and (len(ulabels) >= column)):  #select nth label if possible.
			plt.legend ([ulabels[column]])	

else:			#Show everything on one graph.
	plt.title ("Data and Linear Regression [ " + infilename + ", column " + str(column) + " ]" if (title == False) else title)
	ax2 = ax1.twinx()	#This way gets the same colour cycle for the data and the regression lines
	plt.ylabel("Regression (V)" if (ylabel == False) else ylabel)
	if (column == "all"):	#without error-bars: we don't want them, because it would make this really crowded.
#		the markevery property isn't supported in older versions of scipy.
#		plt.plot(times, (regression), linewidth=1, linestyle='--', marker='^', markevery=(0,n/20), markeredgewidth=1.5)  #Use marker to make regression lines clearer. Also: "markerfacecolor='k'" ?
		plt.plot(times, (regression), linewidth=1, linestyle='--', marker='^',  markeredgewidth=1.5)  #Use marker to make regression lines clearer. Also: "markerfacecolor='k'" ?
	else:
#		plt.errorbar(times, (regression), yerr=s, linewidth=1, linestyle='--', marker='^', markevery=(0,n/20), markeredgewidth=1.5, ecolor='lightgrey')
		plt.errorbar(times, (regression), yerr=s, linewidth=1, linestyle='--', marker='^', markeredgewidth=1.5, ecolor='lightgrey')
	plt.ylim(ax1_ylim)


#Show summary info, print variables.
print "Loaded column " + str(column) + " from file '" + infilename + "': " + str(num_samples) + " points at " + str(samplerate) + " Hz sample-rate, for " + str(duration) + " seconds."
print "Regression is:  y = a + b.x + noise().  Noise has std-dev s. Standard-errors are se(a) and se(b). Pearson's PMCC is r. Dy is (y_n - y_0), b*Dx is b*(x_n - x_0).\n"
print "n:     ", n			#num samples
print "a:     ", a			#a-hat intercept estimator
print "b:     ", b			#b-hat gradient estimator
print "s:     ", s			#noise std-dev (i.e. "amplitude" of gaussian noise)
print "se(a): ", se_a			#std err in estimate of a
print "se(b): ", se_b			#std err in estimate of b
print "r:     ", r			#estimator of pearson's rho. -1 or +1 means strong correlation
print "Dy:    ", delta_y		#Actual value of change in Y (nth sample - 0th sample)
print "b*Dx:  ", b_delta_x		#Best estimate for the "real" change in Y.
print ""

#Now save and show the plots, and exit.
if (outplot):
	plt.savefig (outplot)
	print "Saved file as: " + outplot
if (noplot == False):
	plt.show()
exit (0)
